/**
 * Sencha GXT 4.0.0 - Sencha for GWT
 * Copyright (c) 2006-2015, Sencha Inc.
 *
 * licensing@sencha.com
 * http://www.sencha.com/products/gxt/license/
 *
 * ================================================================================
 * Open Source License
 * ================================================================================
 * This version of Sencha GXT is licensed under the terms of the Open Source GPL v3
 * license. You may use this license only if you are prepared to distribute and
 * share the source code of your application under the GPL v3 license:
 * http://www.gnu.org/licenses/gpl.html
 *
 * If you are NOT prepared to distribute and share the source code of your
 * application under the GPL v3 license, other commercial and oem licenses
 * are available for an alternate download of Sencha GXT.
 *
 * Please see the Sencha GXT Licensing page at:
 * http://www.sencha.com/products/gxt/license/
 *
 * For clarification or additional options, please contact:
 * licensing@sencha.com
 * ================================================================================
 *
 *
 * ================================================================================
 * Disclaimer
 * ================================================================================
 * THIS SOFTWARE IS DISTRIBUTED "AS-IS" WITHOUT ANY WARRANTIES, CONDITIONS AND
 * REPRESENTATIONS WHETHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
 * IMPLIED WARRANTIES AND CONDITIONS OF MERCHANTABILITY, MERCHANTABLE QUALITY,
 * FITNESS FOR A PARTICULAR PURPOSE, DURABILITY, NON-INFRINGEMENT, PERFORMANCE AND
 * THOSE ARISING BY STATUTE OR FROM CUSTOM OR USAGE OF TRADE OR COURSE OF DEALING.
 * ================================================================================
 */
package com.sencha.gxt.themebuilder.config.shared;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class CfgObject extends AbstractCfgObject {
  private final CfgObject parent;
  private final Map<String, AbstractCfgObject> map = new HashMap<String, AbstractCfgObject>();
  private final Map<String, String> remarks = new HashMap<String, String>();

  public CfgObject(SourceInfo source, CfgObject parent) {
    super(source);
    this.parent = parent;
  }

  /**
   * Removes the given key's value from this object. Cannot descend into subobjects or walk up to parent scopes.
   * @param key
   */
  public void erase(String key) {
    assert !key.contains(".") : "Cannot erase nested keys";
    assert map.containsKey(key) : "Cannot remove keys that don't exist";
    map.remove(key);
  }

  public void put(String key, int value) {
    put(key, value, null);
  }

  public void put(String key, int value, String remark) {
    put(key, new CfgIntLiteral(makeSubSource(key), value), remark);
  }


  public void put(String key, double value) {
    put(key, value, null);
  }

  public void put(String key, double value, String remark) {
    put(key, new CfgDoubleLiteral(makeSubSource(key), value), remark);
  }

  public void put(String key, String value) {
    put(key, value, null);
  }

  public void put(String key, String value, String remark) {
    put(key, new CfgStringLiteral(makeSubSource(key), value), remark);
  }

  public void put(String key, AbstractCfgObject value) {
    put (key, value, null);
  }

  public void put(String key, AbstractCfgObject value, String remark) {
    put(Arrays.asList(key.split("\\.")), value, remark);
  }

  public void put(List<String> key, AbstractCfgObject value) {
    put(key, value, null);
  }

  public void put(List<String> key, AbstractCfgObject value, String remark) {
    assert value != null : "Value cannot be empty";
    assert key.size() > 0 : "Key cannot be empty";
    if (key.size() == 1) {
      if (map.containsKey(key.get(0))) {
        throw new IllegalStateException("Property already exists, cannot be redefined: "
                + key + " in " + getSource().getPath() + " cannot be assigned " + value);
      }
      map.put(key.get(0), value);
      if (remark != null) {
        remarks.put(key.get(0), remark);
      }
    } else {
      //TODO not quite right, we should build objects as we go instead of failing our way up
      List<String> subkeys = key.subList(0, key.size() - 1);
      SourceInfo newSource = new SourceInfo(value.getSource().getBeginLine(), value.getSource().getBeginColumn(), value.getSource() + join(subkeys));
      CfgObject object = getOrCreate(newSource, subkeys);

      CfgObject cfgObject = object.resolveObject();
      String lastKey = key.get(key.size() - 1);
      if (cfgObject.map.containsKey(lastKey)) {
        throw new IllegalStateException("Property already exists, cannot be redefined: "
                + key + " in " + getSource().getPath() + " cannot be assigned " + value);
      }
      cfgObject.map.put(lastKey, value);
      if (remark != null) {
        cfgObject.remarks.put(lastKey, remark);
      }
    }
  }

  public AbstractCfgObject get(String key) {
    return get(Arrays.asList(key.split("\\.")));
  }

  public AbstractCfgObject get(List<String> key) {
    assert key.size() > 0 : "Key cannot be empty";
    CfgObject m = this;
    List<String> k = key.subList(0, key.size() - 1);
    while (!k.isEmpty()) {
      try {
        if (!m.map.containsKey(k.get(0))) {
          //no such key here, try parent
          m = null;
          break;
        }
        //using resolveObject to deal with possible references
        m = m.map.get(k.get(0)).resolveObject();
      } catch (WrongTypeException e) {
        //isn't an object, try parent
        m = null;
        break;
      }
      k = k.subList(1, k.size());
    }
    if (m != null) {
      AbstractCfgObject value = m.map.get(key.get(key.size() - 1));
      if (value != null) {
        return value;
      }
    }
    //failed locally, try parent
    if (parent == null) {
      //no parent, throw an exception
      throw new KeyNotFoundException("Can't resolve " + key);
    }
    try {
      return parent.get(key);
    } catch (KeyNotFoundException ex) {
      //since we caught an exception, some nexted code failed to look up its part of the key - wrap and rethrow
      //wrap to part that failed, throw new to show the full key passed int
      throw new KeyNotFoundException("Can't resolve " + key, ex);
    }
  }

  public String getRemark(String key) {
    return remarks.get(key);
  }


  public Set<String> keySet() {
    return map.keySet();
  }

  @Override
  public int resolveInt() {
    throw new WrongTypeException("Object can't be used as an int " + getSource());
  }

  @Override
  public double resolveDouble() {
    throw new WrongTypeException("Object can't be used as a double " + getSource());
  }

  @Override
  public String resolveString() {
    throw new WrongTypeException("Object can't be used as a String " + getSource());
  }

  @Override
  public CfgObject resolveObject() {
    return this;
  }

  @Override
  public CfgArray resolveArray() {
    throw new WrongTypeException("Object can't be used as an Array " + getSource());
  }

  private CfgObject getOrCreate(SourceInfo source, List<String> key) {
    assert key.size() > 0;
    CfgObject obj = this;
    for (String s : key) {
      CfgObject next = (CfgObject) obj.map.get(s);
      if (next == null) {
        next = new CfgObject(source, obj);
        obj.map.put(s, next);
      }
      obj = next;
    }
    return obj;
  }

  private String join(List<String> key) {
    StringBuilder sb = new StringBuilder();
    for (String k : key) {
      assert k != null && k.length() != 0;
      if (sb.length() != 0) {
        sb.append(".");
      }
      sb.append(k);
    }
    return sb.toString();
  }

  private SourceInfo makeSubSource(String key) {
    return new SourceInfo(getSource().getBeginLine(), getSource().getBeginColumn(), getSource().getPath() + key);
  }

  @Override
  public void accept(CfgVisitor visitor) {
    visitor.visitObject(this);
  }
}
